@@ -7,9 +7,10 @@ if version_info.major == 3 and version_info.minor < 6:
     exit(1)
 
 import argparse
+import pprint
 import shlex
 from os.path import abspath
-from subprocess import check_output
+from subprocess import STDOUT, CalledProcessError, check_output
 from typing import Dict
 
 import uvicorn  # type: ignore
@@ -27,9 +28,10 @@ USE_LOCAL = False
 PORT = 8000
 INFO = "SoCo-CLI HTTP API Server v" + version
 PREFIX = "SoCo-CLI: "
+PREFIX_MACRO = PREFIX + "Macro: "
 MACROS: Dict[str, str] = {}
 MACRO_FILE = ""
-
+PP = pprint.PrettyPrinter(indent=len(PREFIX_MACRO))
 
 sc_app = FastAPI()
 
@@ -111,44 +113,56 @@ def macros() -> Dict:
     return MACROS
 
 
+@sc_app.get("/macros/list")
+def macros_list() -> Dict:
+    return MACROS
+
+
+@sc_app.get("/macros/reload")
+def macros_reload() -> Dict:
+    global MACROS
+    _load_macros(MACROS, filename=MACRO_FILE)
+    return MACROS
+
+
 @sc_app.get("/macro/{macro_name}")
 def run_macro(macro_name: str) -> Dict:
-    result = _process_macro(macro_name)
-    return {"result": result}
+    command, result = _process_macro(macro_name)
+    return {"command": command, "result": result}
 
 
 @sc_app.get("/macro/{macro_name}/{arg_1}")
 def run_macro_1(macro_name: str, arg_1: str) -> Dict:
-    result = _process_macro(macro_name, arg_1)
-    return {"result": result}
+    command, result = _process_macro(macro_name, arg_1)
+    return {"command": command, "result": result}
 
 
 @sc_app.get("/macro/{macro_name}/{arg_1}/{arg_2}")
 def run_macro_2(macro_name: str, arg_1: str, arg_2: str) -> Dict:
-    result = _process_macro(macro_name, arg_1, arg_2)
-    return {"result": result}
+    command, result = _process_macro(macro_name, arg_1, arg_2)
+    return {"command": command, "result": result}
 
 
 @sc_app.get("/macro/{macro_name}/{arg_1}/{arg_2}/{arg_3}")
 def run_macro_3(macro_name: str, arg_1: str, arg_2: str, arg_3: str) -> Dict:
-    result = _process_macro(macro_name, arg_1, arg_2, arg_3)
-    return {"result": result}
+    command, result = _process_macro(macro_name, arg_1, arg_2, arg_3)
+    return {"command": command, "result": result}
 
 
 @sc_app.get("/macro/{macro_name}/{arg_1}/{arg_2}/{arg_3}/{arg_4}")
 def run_macro_4(
     macro_name: str, arg_1: str, arg_2: str, arg_3: str, arg_4: str
 ) -> Dict:
-    result = _process_macro(macro_name, arg_1, arg_2, arg_3, arg_4)
-    return {"result": result}
+    command, result = _process_macro(macro_name, arg_1, arg_2, arg_3, arg_4)
+    return {"command": command, "result": result}
 
 
 @sc_app.get("/macro/{macro_name}/{arg_1}/{arg_2}/{arg_3}/{arg_4}/{arg_5}")
 def run_macro_5(
     macro_name: str, arg_1: str, arg_2: str, arg_3: str, arg_4: str, arg_5: str
 ) -> Dict:
-    result = _process_macro(macro_name, arg_1, arg_2, arg_3, arg_4, arg_5)
-    return {"result": result}
+    command, result = _process_macro(macro_name, arg_1, arg_2, arg_3, arg_4, arg_5)
+    return {"command": command, "result": result}
 
 
 @sc_app.get("/{speaker}/{action}")
@@ -242,13 +256,14 @@ def main() -> None:
         exit(1)
 
 
-def _process_macro(macro_name: str, *args) -> str:
+def _process_macro(macro_name: str, *args) -> tuple[str, str]:
     # Look up the macro
     try:
         macro = _lookup_macro(macro_name)
+        print(PREFIX_MACRO + "Processing macro '{}' = '{}'".format(macro_name, macro))
     except KeyError:
-        print(PREFIX + "macro '{}' not found".format(macro_name))
-        return "Error: macro '{}' not found".format(macro_name)
+        print(PREFIX_MACRO + "macro '{}' not found".format(macro_name))
+        return "", "Error: macro '{}' not found".format(macro_name)
 
     # Substitute variable arguments
     sonos_command_line = _substitute_variables(macro, args)
@@ -260,12 +275,15 @@ def _process_macro(macro_name: str, *args) -> str:
     sonos_command_line = "sonos " + sonos_command_line
 
     # Execute the command
-    print(PREFIX + "Executing: " + sonos_command_line)
+    print(PREFIX_MACRO + "Executing: '" + sonos_command_line + "' in a subprocess")
     try:
-        output = check_output(sonos_command_line, shell=True)
-        return "Command line output: " + output.decode("utf-8")
-    except:
-        return "Error running command line: " + sonos_command_line
+        output = check_output(sonos_command_line, stderr=STDOUT, shell=True)
+        print(PREFIX_MACRO + "Exit code = 0")
+        return sonos_command_line, output.decode("utf-8").rstrip()
+    except CalledProcessError as exc:
+        error = exc.output.decode("utf-8").rstrip().replace("\n", "; ")
+        print(PREFIX_MACRO + "Exit code = {} [{}]".format(exc.returncode, error))
+        return sonos_command_line, error
 
 
 def _lookup_macro(macro_name: str) -> str:
@@ -275,24 +293,60 @@ def _lookup_macro(macro_name: str) -> str:
 
 def _substitute_variables(macro: str, args: tuple) -> str:
     """Substitute positional parameters with supplied variables."""
+    parameters_list = ["%1", "%2", "%3", "%4", "%5"]
+    supplied_parameters = set(parameters_list[: len(args)])
+    parameters = set(parameters_list)
+    used_parameters = []
+    unsatisfied_parameters = set()
+    variables_used = []
+
     elements = shlex.split(macro)
-    sonos_command_line = ""
+    sonos_command_line_terms = []
     for element in elements:
-        if element in ["%1", "%2", "%3", "%4", "%5"]:
+        if element in parameters:
             try:
                 arg_sub = _quote_if_contains_space(args[int(element[1:]) - 1])
-                sonos_command_line = sonos_command_line + " " + arg_sub
+                sonos_command_line_terms.append(arg_sub)
+                used_parameters.append(element)
+                variables_used.append(arg_sub)
             except IndexError:
-                print(
-                    PREFIX
-                    + "Argument substitution failed for argument '{}'".format(element)
-                )
-                # Proceed with unsubstituted element; will probably cause command
-                # failure
-                sonos_command_line = sonos_command_line + " " + element
+                # Omit unsatisfied arguments and continue
+                unsatisfied_parameters.add(element)
         else:
-            sonos_command_line = sonos_command_line + " " + element
-    return sonos_command_line
+            sonos_command_line_terms.append(_quote_if_contains_space(element))
+
+    used_parameters_set = set(used_parameters)
+
+    # Print out parameter usage
+    if len(used_parameters) > 0:
+        print(
+            PREFIX_MACRO
+            + "Parameter variables used: {} -> {}".format(
+                used_parameters, variables_used
+            )
+        )
+    if len(unsatisfied_parameters) > 0:
+        print(
+            PREFIX_MACRO
+            + "Parameter variables not supplied for: {}".format(
+                sorted(list(unsatisfied_parameters))
+            )
+        )
+    if len(supplied_parameters - used_parameters_set) > 0:
+        unused_list = sorted(list(supplied_parameters - used_parameters_set))
+        unused_variables = []
+        for unused in unused_list:
+            unused_variables.append(_quote_if_contains_space(args[int(unused[1:]) - 1]))
+        print(
+            PREFIX_MACRO
+            + "Parameter variables supplied but not used: {} -> {}".format(
+                sorted(list(supplied_parameters - used_parameters_set)),
+                unused_variables,
+            )
+        )
+
+    # Return the substituted command line
+    return " ".join(sonos_command_line_terms)
 
 
 def _substitute_speaker_ips(macro: str, use_local: bool = False) -> str:
@@ -300,36 +354,45 @@ def _substitute_speaker_ips(macro: str, use_local: bool = False) -> str:
     Substitute speaker names for IP addresses, for efficiency.
     Speaker names must be exact.
     """
-    terms = shlex.split(macro)
+    elements = shlex.split(macro)
     new_macro_list = []
-    for term in terms:
-        device, error_msg = get_speaker(term, use_local_speaker_list=use_local)
-        if device is not None and device.player_name == term:
+    for element in elements:
+        device, error_msg = get_speaker(element, use_local_speaker_list=use_local)
+        if device is not None and device.player_name == element:
             new_macro_list.append(device.ip_address)
+            print(
+                PREFIX_MACRO
+                + "Substituting speaker name '{}' by IP address '{}'".format(
+                    device.player_name, device.ip_address
+                )
+            )
         else:
-            new_macro_list.append(_quote_if_contains_space(term))
-    new_macro = " ".join(new_macro_list)
-    return new_macro
+            new_macro_list.append(_quote_if_contains_space(element))
+    return " ".join(new_macro_list)
 
 
 def _load_macros(macros: dict, filename: str) -> bool:
-    print(PREFIX + "Attempting to load macros from '{}'".format(filename))
+    print(PREFIX_MACRO + "Attempting to (re)load macros from '{}'".format(filename))
     try:
         with open(filename, "r") as f:
             line = f.readline()
             while line != "":
                 if not line.startswith("#") and line != "\n":
                     if line.count("=") != 1:
-                        print(PREFIX + "Malformed macro '{}'... ignored".format(line))
+                        print(
+                            PREFIX_MACRO
+                            + "Malformed macro '{}'... ignored".format(line)
+                        )
                         print(line, end="")
                     else:
                         macro = line.split("=")
                         macros[macro[0].strip()] = macro[1].strip()
                 line = f.readline()
-        print(PREFIX + "Loaded macros: {}".format(macros))
+        print(PREFIX_MACRO + "Loaded macros:")
+        PP.pprint(macros)
         return True
     except:
-        print(PREFIX + "Macro file not found")
+        print(PREFIX_MACRO + "Macro file not found")
         return False
 
 
